#define ARDBUFFER 101
#include <LedControl.h>
#include <LibPrintf.h>

/* 
 * Ultrasonic sensors.
 */
#define NO_SENSORS 2

const int ECHO_PINS[NO_SENSORS] = {12, 10};
const int TRIGGER_PINS[NO_SENSORS] = {13, 11};
int distance[NO_SENSORS], tmp[NO_SENSORS];

/* 
 * LED matrix.
 */
const int DIN_PIN = 9;
const int CS_PIN = 8;
const int CLK_PIN = 7;
LedControl display = LedControl(DIN_PIN, CLK_PIN, CS_PIN);

/*
 * python_connected = true only when Python was successfully connected on the serial port.
 * start = true only when Python was connected and the user made the starting gesture.
 */
bool python_connected = false, start = false;

/*
 * Gesture icons.
 */
byte Hi[8] = {
  B00000000,
  B00000000,
  B00101000,
  B00101010,
  B00111000,
  B00101010,
  B00101010,
  B00000000
};
byte loading[8] = {
  B00000000,
  B01111110,
  B00111100,
  B00011000,
  B00011000,
  B00111100,
  B01111110,
  B00000000
};
byte play_pause[8] = {
  B00000000,
  B10000000,
  B11001010,
  B11101010,
  B11101010,
  B11001010,
  B10000000,
  B00000000
};
byte volume_up[8] = {
  B00000000,
  B00011000,
  B00111100,
  B01111110,
  B00011000,
  B00011000,
  B00011000,
  B00000000
};
byte volume_down[8] = {
  B00000000,
  B00011000,
  B00011000,
  B00011000,
  B01111110,
  B00111100,
  B00011000,
  B00000000
};
byte swipe_left[8] = {
  B00000000,
  B00010000,
  B00110000,
  B01111110,
  B01111110,
  B00110000,
  B00010000,
  B00000000
};
byte swipe_right[8] = {
  B00000000,
  B00001000,
  B00001100,
  B01111110,
  B01111110,
  B00001100,
  B00001000,
  B00000000
};

/*
 * Finds the distance measured in centimeters by the i-th sensor having:
 * - echo pin = ECHO_PINS[i]
 * - trigger pin = TRIGGER_PIN[i]
 */
void find_distance(int i)
{
  // Makes sure that the trigger pin is clear by setting it to a LOW state for 2 microseconds.
  digitalWrite(TRIGGER_PINS[i], LOW);
  delayMicroseconds(2);

  // Generates the ultrasound wave by setting it to a HIGH state for 10 microseconds.
  digitalWrite(TRIGGER_PINS[i], HIGH);
  delayMicroseconds(10);

  // Resets the trigger pin.
  digitalWrite(TRIGGER_PINS[i], LOW);

  // Reads the echo pin and returns the sound wave travel time in microseconds.
  long duration = pulseIn(ECHO_PINS[i], HIGH);

  /*
   * Computes the distance knowing:
   * - the speed of sound = 340m/s
   * - the sound wave travels both forward and backwards
   *   => the actual distance is only half of the distance calculated.
   * - the duration is measured in microseconds
   *   => duration(s) = 10^(-6) * duration(us)
   * - the resulted distance should be computed in centimeters
   *   => distance(m) = 10^(-2) * distance(cm)
   * 
   * 10^(-2) * distance(cm) = (speed * 10^(-6) * duration(us)) / 2
   * => distance(cm) = (340 * 10^(-4) / 2) * duration(us) = 0.017 * duration(us)
   */
  distance[i] = 0.017 * duration;
}

/*
 * Finds the distance for each sensor.
 */
void find_distances(void)
{
  for (int i = 0; i < NO_SENSORS; i++) {
    find_distance(i);
  }
}

/*
 * Displays an image on the LED matrix.
 * The image is an array of bytes where each byte represents one line on the matrix.
 */
void display_image(byte image[8])
{
  for (int i = 0; i < 8; i++) {
    display.setRow(0, i, image[i]);
  }
  delay(100);
}

/*
 * Waits until the connection with Python is established.
 */
void wait_begin()
{
  while (!Serial.available());

  String msg = Serial.readString();
  if (msg.equals("start")) {
    python_connected = true;
  }
}

void setup() 
{
  // Sets the baud rate.
  Serial.begin(9600);

  for (int i = 0; i < NO_SENSORS; i++) {
    pinMode(ECHO_PINS[i], INPUT);
    pinMode(TRIGGER_PINS[i], OUTPUT);
  }

  // Sets up the LED matrix.
  display.clearDisplay(0);
  display.shutdown(0, false);
  display.setIntensity(0, 10);

  wait_begin();
}

/*
 * Finds the current gesture and sends it to Python.
 */
void find_gesture()
{ 
  // Returns if Python is not connected yet.
  if (!python_connected) {
    return;
  }

  find_distances();

  /*
   * Both hands should be placed in front of the sensors to start
   * the gesture recognition.
   */
  if (!start) {
    if (distance[0] < 30 && distance[1] < 30) {
      display_image(Hi);
      start = true;
    }
    delay(3000);
    return;
  }

  display_image(loading);

  // Finds the gesture and sends it to Python.
  if (distance[0] < 30 && distance[1] < 30) {
    display_image(play_pause);
    printf("play/pause\n");
  } else {
    memcpy(tmp, distance, NO_SENSORS * sizeof(int));
    
    delay(500);
    find_distances();
    
    if (tmp[0] < 30) {
      if (distance[0] < 30) {
        printf("volume down\n");
        display_image(volume_down);
      } else {
        printf("swipe left\n");
        display_image(swipe_left);
      }
    } else if (tmp[1] < 30) {
      if (distance[1] < 30) {
        printf("volume up\n");
        display_image(volume_up);
      } else {
        printf("swipe right\n");
        display_image(swipe_right);
      }
    } else {
      return;
    }
  }
  delay(3000);
}

void loop()
{
  find_gesture();
}
